#include <InternalDisplay.hpp>


void InternalDisplay::begin(ESPMegaIoT *iot, std::function<rtctime_t()> getRtcTime) {
    ESP_LOGD("InternalDisplay", "Assigning IoT Module and RTC Time Getter");
    this->iot = iot;
    this->getRtcTime = getRtcTime;
    ESP_LOGD("InternalDisplay", "Acquiring Network and MQTT Configs");
    this->mqttConfig = this->iot->getMqttConfig();
    this->networkConfig = this->iot->getNetworkConfig();
    // Register callbacks
    ESP_LOGD("InternalDisplay", "Registering Callbacks");
    auto bindedPageChangeCallback = std::bind(&InternalDisplay::handlePageChange, this, std::placeholders::_1);
    this->registerPageChangeCallback(bindedPageChangeCallback);
    auto bindedTouchCallback = std::bind(&InternalDisplay::handleTouch, this, std::placeholders::_1, std::placeholders::_2, std::placeholders::_3);
    this->registerTouchCallback(bindedTouchCallback);
    ESP_LOGD("InternalDisplay", "Binding Page Change Callback");
    auto bindedInputStateChangeCallback = std::bind(&InternalDisplay::handleInputStateChange, this, std::placeholders::_1, std::placeholders::_2);
    ESP_LOGD("InternalDisplay", "Binding Input State Change Callback");
    auto bindedPwmStateChangeCallback = std::bind(&InternalDisplay::handlePwmStateChange, this, std::placeholders::_1, std::placeholders::_2, std::placeholders::_3);
    ESP_LOGD("InternalDisplay", "Registering inputCard Callbacks");
    this->inputCard->registerCallback(bindedInputStateChangeCallback);
    ESP_LOGD("InternalDisplay", "Registering outputCard Callbacks");
    this->outputCard->registerChangeCallback(bindedPwmStateChangeCallback);
    // Initialize the display
    ESP_LOGD("InternalDisplay", "Initializing DisplayAdapter");
    this->displayAdapter->begin(115200);
    ESP_LOGD("InternalDisplay", "Setting DisplayAdapter Timeout");
    this->displayAdapter->setTimeout(100);
    ESP_LOGD("InternalDisplay", "Flushing DisplayAdapter");
    this->displayAdapter->flush();
    ESP_LOGD("InternalDisplay", "Resetting Display");
    this->reset();
    delay(500);
    ESP_LOGD("InternalDisplay", "Jumping to Page 1");
    this->jumpToPage(1);
    ESP_LOGD("InternalDisplay", "Display Initialization Complete");
}


void InternalDisplay::loop() {
    // Keep reading the Serial Adapter
    this->recieveSerialCommand();

    // Refresh the top bar every 5 seconds
    static uint32_t lastTopBarRefresh;
    if (millis() - lastTopBarRefresh > INTERNAL_DISPLAY_TOP_BAR_REFRESH_INTERVAL) {
        this->updateStatusIcons(this->iot->networkConnected(), this->iot->mqttConnected());
        lastTopBarRefresh = millis();
    }
    // Refresh the clock every 10 seconds
    static uint32_t lastClockRefresh;
    if (millis() - lastClockRefresh > INTERNAL_DISPLAY_CLOCK_REFRESH_INTERVAL) {
        this->updateClock();
        lastClockRefresh = millis();
    }
}



void InternalDisplay::handleInputStateChange(uint8_t pin, bool state) {
    // If the input card is binded to the display and the current page is the input page
    // then update the respective input component
    if (this->inputCard!=nullptr || this->currentPage != INTERNAL_DISPLAY_INPUT_PAGE) return;
    // Update the input state
    this->setInputMarker(pin, state);
}

void InternalDisplay::handlePwmStateChange(uint8_t pin, bool state, uint16_t value) {
    // If the output card is binded to the display and the current page is the output page
    // then update the respective output component
    if (this->outputCard!=nullptr || this->currentPage != INTERNAL_DISPLAY_OUTPUT_PAGE) return;
    // Update the output state
    this->setOutputBar(pin, value);
    this->setOutputStateColor(pin, state);
    // Refresh the PWM Adjustment page if the current page is the PWM Adjustment page and the pin is the same
    if (this->currentPage == INTERNAL_DISPLAY_PWM_ADJUSTMENT_PAGE && this->pmwAdjustmentPin == pin) {
        this->refreshPWMAdjustment();
    }
}

void InternalDisplay::handlePageChange(uint8_t page) {
    // Refresh the page
    this->refreshPage(page);
}

void InternalDisplay::saveNetworkConfig() {
    // TODO: implementation
}

void InternalDisplay::saveMQTTConfig() {
    // TODO: implementation
}

void InternalDisplay::updateStatusIcons(bool networkStatus, bool mqttStatus) {
    this->setNumber("server.pic", mqttStatus ? PIC_MQTT_CONNECTED : PIC_MQTT_DISCONNECTED);
    this->setNumber("lan.pic", networkStatus ? PIC_LAN_CONNECTED : PIC_LAN_DISCONNECTED);
}

void InternalDisplay::updateClock() {
    rtctime_t time = this->getRtcTime();
    this->displayAdapter->print("time.txt=");
    this->displayAdapter->print(time.hours%12);
    this->displayAdapter->print(":");
    this->displayAdapter->print(time.minutes);
    this->displayAdapter->print(" ");
    this->displayAdapter->print(time.hours/12 ? "PM" : "AM");
    this->sendStopBytes();

}

void InternalDisplay::refreshPage() {
    this->refreshPage(this->currentPage);
}

void InternalDisplay::refreshPage(uint8_t page) {
    switch (page) {
        case INTERNAL_DISPLAY_DASHBOARD_PAGE:
            this->refreshDashboard();
            break;
        case INTERNAL_DISPLAY_INPUT_PAGE:
            this->refreshInput();
            break;
        case INTERNAL_DISPLAY_OUTPUT_PAGE:
            this->refreshOutput();
            break;
        case INTERNAL_DISPLAY_AC_PAGE:
            if (this->climateCard == nullptr) {
                this->jumpToPage(INTERNAL_DISPLAY_CLIMATE_NULL_PTR_PAGE);
                break;
            }
            this->refreshAC();
            break;
        case INTERNAL_DISPLAY_PWM_ADJUSTMENT_PAGE:
            this->refreshPWMAdjustment();
            break;
        default:
            break;
    }
}

void InternalDisplay::refreshDashboard() {
    // The dashboard have the following components:
    // 1. Hostname
    // 2. IP Address
    // 3. MQTT Server with port
    // 4. MQTT Connection status
    this->setString("hostname.txt", this->networkConfig->hostname);
    // Construct the IP address string
    static char ip_address[25];
    sprintf(ip_address, "%d.%d.%d.%d", this->networkConfig->ip[0], this->networkConfig->ip[1], this->networkConfig->ip[2], this->networkConfig->ip[3]);
    this->setString("ip_address.txt", ip_address);
    // Send the MQTT server and port
    this->displayAdapter->print("server_address.txt=\"");
    this->displayAdapter->print(this->mqttConfig->mqtt_server);
    this->displayAdapter->print("\"");
    this->sendStopBytes();
    // Send the MQTT connection status
    this->setString("status_txt.txt", this->iot->mqttConnected() ? MSG_MQTT_CONNECTED : MSG_MQTT_DISCONNECTED);
}

void InternalDisplay::refreshInput() {
    for (uint8_t i=0; i<16; i++) {
        this->setInputMarker(i, this->inputCard->digitalRead(i, false));
    }
}

void InternalDisplay::refreshOutput() {
    for (uint8_t i=0; i<16; i++) {
        this->setOutputBar(i, this->outputCard->getValue(i));
        this->setOutputStateColor(i, this->outputCard->getState(i));
    }
}

void InternalDisplay::refreshAC() {
    this->displayAdapter->print("temp.txt=\"");
    this->displayAdapter->print(this->climateCard->getTemperature());
    this->displayAdapter->print("C\"");
    this->sendStopBytes();
    this->displayAdapter->print("fan_auto.pic=");
    this->displayAdapter->print(this->climateCard->getFanSpeed() == AC_FAN_SPEED_AUTO ? PIC_AC_FAN_SPEED_AUTO_ACTIVE : PIC_AC_FAN_SPEED_AUTO_INACTIVE);
    this->sendStopBytes();
    this->displayAdapter->print("fan_low.pic=");
    this->displayAdapter->print(this->climateCard->getFanSpeed() == AC_FAN_SPEED_LOW ? PIC_AC_FAN_SPEED_LOW_ACTIVE : PIC_AC_FAN_SPEED_LOW_INACTIVE);
    this->sendStopBytes();
    this->displayAdapter->print("fan_mid.pic=");
    this->displayAdapter->print(this->climateCard->getFanSpeed() == AC_FAN_SPEED_MEDIUM ? PIC_AC_FAN_SPEED_MEDIUM_ACTIVE : PIC_AC_FAN_SPEED_MEDIUM_INACTIVE);
    this->sendStopBytes();
    this->displayAdapter->print("fan_high.pic=");
    this->displayAdapter->print(this->climateCard->getFanSpeed() == AC_FAN_SPEED_HIGH ? PIC_AC_FAN_SPEED_HIGH_ACTIVE : PIC_AC_FAN_SPEED_HIGH_INACTIVE);
    this->sendStopBytes();
    this->displayAdapter->print("mode_off.pic=");
    this->displayAdapter->print(this->climateCard->getMode() == AC_MODE_OFF ? PIC_AC_MODE_OFF_ACTIVE : PIC_AC_MODE_OFF_INACTIVE);
    this->sendStopBytes();
    this->displayAdapter->print("mode_fan.pic=");
    this->displayAdapter->print(this->climateCard->getMode() == AC_MODE_FAN_ONLY ? PIC_AC_MODE_FAN_ACTIVE : PIC_AC_MODE_FAN_INACTIVE);
    this->sendStopBytes();
    this->displayAdapter->print("mode_cool.pic=");
    this->displayAdapter->print(this->climateCard->getMode() == AC_MODE_COOL ? PIC_AC_MODE_COOL_ACTIVE : PIC_AC_MODE_COOL_INACTIVE);
    this->sendStopBytes();
    if (this->climateCard->getSensorType() == AC_SENSOR_TYPE_DHT22) {
        this->displayAdapter->print("roomtemp.txt=\"");
        this->displayAdapter->print(this->climateCard->getRoomTemperature());
        this->displayAdapter->print("C\"");
        this->sendStopBytes();
        this->displayAdapter->print("roomhumid.txt=\"");
        this->displayAdapter->print(this->climateCard->getHumidity());
        this->displayAdapter->print("%\"");
        this->sendStopBytes();
    }
    else if(this->climateCard->getSensorType() == AC_SENSOR_TYPE_DS18B20) {
        this->displayAdapter->print("roomtemp.txt=\"");
        this->displayAdapter->print(this->climateCard->getRoomTemperature());
        this->displayAdapter->print("C\"");
        this->sendStopBytes();
        this->setString("roomhumid.txt", "N/A");
    }
    else {
        this->setString("roomtemp.txt", "N/A");
        this->setString("roomhumid.txt", "N/A");
    }
}

void InternalDisplay::setOutputBar(uint8_t pin, uint16_t value) {
    // Write the value to the output bar
    this->displayAdapter->print("j");
    this->displayAdapter->print(pin);
    this->displayAdapter->print(".val=");
    this->displayAdapter->print((int)(value*100/4095));
    this->sendStopBytes();
}

void InternalDisplay::setOutputStateColor(uint8_t pin, bool state) {
    this->displayAdapter->print("j");
    this->displayAdapter->print(pin);
    this->displayAdapter->print(".ppic=");
    this->displayAdapter->print(state ? PIC_PWM_BAR_ON : PIC_PWM_BAR_OFF);
    this->sendStopBytes();

}

void InternalDisplay::setInputMarker(uint8_t pin, bool state) {
    this->displayAdapter->print("I");
    this->displayAdapter->print(pin);
    this->displayAdapter->print(".val=");
    this->displayAdapter->print(state ? 0:1);
    this->sendStopBytes();
}

InternalDisplay::InternalDisplay(HardwareSerial *displayAdapter) : ESPMegaDisplay(displayAdapter) {
    this->currentPage = INTERNAL_DISPLAY_DASHBOARD_PAGE;
    this->iot = nullptr;
    this->inputCard = nullptr;
    this->outputCard = nullptr;
    this->climateCard = nullptr;
    this->pmwAdjustmentPin = 0;
}

void InternalDisplay::bindInputCard(DigitalInputCard *inputCard) {
    this->inputCard = inputCard;
}

void InternalDisplay::bindOutputCard(DigitalOutputCard *outputCard) {
    this->outputCard = outputCard;
}

// This assume that your ClimeateCard has the mode and fan speed names in the following order:
// mode: [off, fan_only, cool]
// fan_speed: [auto, low, medium, high]
void InternalDisplay::bindClimateCard(ClimateCard *climateCard) {
    this->climateCard = climateCard;
}

void InternalDisplay::refreshPWMAdjustment() {
    // The PWM Adjustment page have the following components:
    // pwm_value -> a slider to adjust the PWM value
    // pwm_state -> a button to toggle the PWM state
    // pwm_id -> a text to show the PWM pin

    // Refresh the PWM pin
    this->refreshPWMAdjustmentId();
    // Refresh the PWM value
    this->refreshPWMAdjustmentSlider();
    // Refresh the PWM state
    this->refreshPWMAdjustmentState();
}

void InternalDisplay::refreshPWMAdjustmentId() {
    // Send the PWM pin
    this->displayAdapter->print("pwm_id.txt=\"P");
    this->displayAdapter->print(pmwAdjustmentPin);
    this->displayAdapter->print("\"");
    this->sendStopBytes();
}

void InternalDisplay::refreshPWMAdjustmentSlider() {
    // Send the PWM value
    this->displayAdapter->print("pwm_value.val=");
    this->displayAdapter->print(this->outputCard->getValue(this->pmwAdjustmentPin));
    this->sendStopBytes();
}

void InternalDisplay::refreshPWMAdjustmentState() {
    // Send the PWM state
    this->displayAdapter->print("pwm_state.txt=\"");
    this->displayAdapter->print(this->outputCard->getState(this->pmwAdjustmentPin) ? MSG_PWM_ADJUSTMENT_STATE_ON : MSG_PWM_ADJUSTMENT_STATE_OFF);
    this->displayAdapter->print("\"");
    this->sendStopBytes();
}

void InternalDisplay::handleTouch(uint8_t page, uint8_t component, uint8_t type) {
    // Switch based on the page
    switch (page) {
        case INTERNAL_DISPLAY_AC_PAGE:
            this->handleACTouch(type, component);
            break;
        case INTERNAL_DISPLAY_PWM_ADJUSTMENT_PAGE:
            this->handlePWMAdjustmentTouch(type, component);
            break;
        default:
            break;
    }
}

void InternalDisplay::handleACTouch(uint8_t type, uint8_t component) {
    // b1 [component 18] -> inclement AC temperature by 1
    // b0 [component 17] -> declement AC temperature by 1
    // fan_auto [component 4] -> set the fan speed to auto
    // fan_low [component 5] -> set the fan speed to low
    // fan_med [component 6] -> set the fan speed to medium
    // fan_high [component 7] -> set the fan speed to high
    // mode_off [component 10] -> set the mode to off
    // mode_fan [component 9] -> set the mode to fan only
    // mode_cool [component 8] -> set the mode to cool

    // For b0 and b1, if the type is not release then return
    // For other components, if the type is not press then return
    if ((component == 17 || component == 18) && type != TOUCH_TYPE_RELEASE) return;
    if ((component != 17 && component != 18) && type != TOUCH_TYPE_PRESS) return;

    // Switch based on the component
    switch (component) {
        case 17:
            // Decrement the temperature
            this->climateCard->setTemperature(this->climateCard->getTemperature() - 1);
            break;
        case 18:
            // Increment the temperature
            this->climateCard->setTemperature(this->climateCard->getTemperature() + 1);
            break;
        case 4:
            // Set the fan speed to auto
            this->climateCard->setFanSpeed(AC_FAN_SPEED_AUTO);
            break;
        case 5:
            // Set the fan speed to low
            this->climateCard->setFanSpeed(AC_FAN_SPEED_LOW);
            break;
        case 6:
            // Set the fan speed to medium
            this->climateCard->setFanSpeed(AC_FAN_SPEED_MEDIUM);
            break;
        case 7:
            // Set the fan speed to high
            this->climateCard->setFanSpeed(AC_FAN_SPEED_HIGH);
            break;
        case 10:
            // Set the mode to off
            this->climateCard->setMode(AC_MODE_OFF);
            break;
        case 9:
            // Set the mode to fan only
            this->climateCard->setMode(AC_MODE_FAN_ONLY);
            break;
        case 8:
            // Set the mode to cool
            this->climateCard->setMode(AC_MODE_COOL);
            break;
        default:
            break;
    }
    // Refresh the AC page
    this->refreshAC();
}

void InternalDisplay::handlePWMAdjustmentTouch(uint8_t type, uint8_t component) {
    // b0 [component 5] -> decrement the PWM id if its greater than 0, else set it to 15
    // b1  [component 6] -> increment the PWM id if its less than 15, else set it to 0
    // pwm_state [component 4] -> toggle the PWM state
    // pwm_value [component 1] -> set the PWM value based on the slider value
    // If the type is not release then return
    if (type != TOUCH_TYPE_RELEASE) return;
    uint16_t val = 0;
    // switch based on the component
    switch (component) {
        case 5:
            // Decrement the PWM id
            this->pmwAdjustmentPin = this->pmwAdjustmentPin > 0 ? this->pmwAdjustmentPin - 1 : 15;
            this->refreshPWMAdjustment();
            break;
        case 6:
            // Increment the PWM id
            this->pmwAdjustmentPin = this->pmwAdjustmentPin < 15 ? this->pmwAdjustmentPin + 1 : 0;
            this->refreshPWMAdjustment();
            break;
        case 4:
            // Toggle the PWM state
            this->outputCard->setState(this->pmwAdjustmentPin, !this->outputCard->getState(this->pmwAdjustmentPin));
            this->refreshPWMAdjustmentState();
            break;
        case 1:
            // Set the PWM value
            val = (uint16_t)this -> getNumber("pwm_value.val");
            this->outputCard->setValue(this->pmwAdjustmentPin, val);
            break;
        default:
            break;
    }
    


}